<p class="logo">
  <a href="{{ pathto(master_doc) }}">
    <h2>Flask-SignalBus</h2>
  </a>
</p>

<p>
  <iframe src="https://ghbtns.com/github-btn.html?user=epandurski&repo=flask_signalbus&type=star&count=true&size=large"
          allowtransparency="true" frameborder="0" scrolling="0" width="200px" height="35px" style="margin-left: 32px"></iframe>
  <br/>
</p>
